@@ -3,32 +3,32 @@
 """
 
 import sys
+from typing import Tuple, List, Union
 
 import numpy as np
-from tensorflow.keras.layers import ActivityRegularization
-from tensorflow.keras.layers import AdditiveAttention
-from tensorflow.keras.layers import Attention
-from tensorflow.keras.layers import BatchNormalization, Concatenate, LayerNormalization, UnitNormalization
-from tensorflow.keras.layers import Bidirectional, TimeDistributed
-from tensorflow.keras.layers import CategoryEncoding
-from tensorflow.keras.layers import Conv1D, ZeroPadding1D, Cropping1D
-from tensorflow.keras.layers import Conv2D, ZeroPadding2D, Cropping2D, CenterCrop
-from tensorflow.keras.layers import Embedding, Normalization, Rescaling, Resizing
-from tensorflow.keras.layers import GlobalAveragePooling1D, GlobalMaxPooling1D
-from tensorflow.keras.layers import GlobalAveragePooling2D, GlobalMaxPooling2D
-from tensorflow.keras.layers import GlobalAveragePooling3D, GlobalMaxPooling3D
-from tensorflow.keras.layers import Input, Dense, Dropout, Flatten, Activation
-from tensorflow.keras.layers import LSTM, GRU
-from tensorflow.keras.layers import LeakyReLU, ELU, PReLU, ReLU
-from tensorflow.keras.layers import MaxPooling1D, AveragePooling1D, UpSampling1D
-from tensorflow.keras.layers import MaxPooling2D, AveragePooling2D, UpSampling2D
-from tensorflow.keras.layers import MaxPooling3D, AveragePooling3D
-from tensorflow.keras.layers import MultiHeadAttention
-from tensorflow.keras.layers import Multiply, Add, Subtract, Average, Maximum, Minimum, Dot
-from tensorflow.keras.layers import Permute, Reshape, RepeatVector
-from tensorflow.keras.layers import SeparableConv2D, DepthwiseConv2D
-from tensorflow.keras.layers import ZeroPadding3D, Cropping3D
-from tensorflow.keras.models import Model, load_model, Sequential
+from keras.layers import ActivityRegularization
+from keras.layers import AdditiveAttention
+from keras.layers import Attention
+from keras.layers import BatchNormalization, Concatenate, LayerNormalization, UnitNormalization
+from keras.layers import CategoryEncoding, Embedding
+from keras.layers import Conv1D, ZeroPadding1D, Cropping1D
+from keras.layers import Conv2D, ZeroPadding2D, Cropping2D, CenterCrop
+from keras.layers import GlobalAveragePooling1D, GlobalMaxPooling1D
+from keras.layers import GlobalAveragePooling2D, GlobalMaxPooling2D
+from keras.layers import GlobalAveragePooling3D, GlobalMaxPooling3D
+from keras.layers import Identity, Conv2DTranspose, Conv1DTranspose
+from keras.layers import Input, Dense, Dropout, Flatten, Activation
+from keras.layers import LeakyReLU, ELU, PReLU, ReLU
+from keras.layers import MaxPooling1D, AveragePooling1D, UpSampling1D
+from keras.layers import MaxPooling2D, AveragePooling2D, UpSampling2D
+from keras.layers import MaxPooling3D, AveragePooling3D
+from keras.layers import MultiHeadAttention
+from keras.layers import Multiply, Add, Subtract, Average, Maximum, Minimum
+from keras.layers import Normalization, Rescaling, Resizing
+from keras.layers import Permute, Reshape, RepeatVector
+from keras.layers import SeparableConv2D, DepthwiseConv2D
+from keras.layers import ZeroPadding3D, Cropping3D
+from keras.models import Model, load_model, Sequential
 
 __author__ = "Tobias Hermann"
 __copyright__ = "Copyright 2017, Tobias Hermann"
@@ -36,13 +36,18 @@ __license__ = "MIT"
 __maintainer__ = "Tobias Hermann, https://github.com/Dobiasd/frugally-deep"
 __email__ = "editgym@gmail.com"
 
+NDFloat32Array = np.typing.NDArray[np.float32]
+NDUInt32Array = np.typing.NDArray[np.int32]
+Shape = Tuple[int, ...]
+VariableShape = Tuple[Union[None, int], ...]
 
-def replace_none_with(value, shape):
+
+def replace_none_with(value: int, shape: VariableShape) -> Shape:
     """Replace every None with a fixed value."""
     return tuple(list(map(lambda x: x if x is not None else value, shape)))
 
 
-def get_shape_for_random_data(data_size, shape):
+def get_shape_for_random_data(data_size: int, shape: Shape) -> Shape:
     """Include size of data to generate into shape."""
     if len(shape) == 5:
         return (data_size, shape[0], shape[1], shape[2], shape[3], shape[4])
@@ -57,46 +62,47 @@ def get_shape_for_random_data(data_size, shape):
     raise ValueError('can not use shape for random data:', shape)
 
 
-def generate_random_data(data_size, shape):
+def generate_random_data(data_size: int, shape: VariableShape) -> NDFloat32Array:
     """Random data for training."""
     return np.random.random(
-        size=get_shape_for_random_data(data_size, replace_none_with(42, shape)))
+        size=get_shape_for_random_data(data_size, replace_none_with(42, shape))).astype(np.float32)
 
 
-def generate_input_data(data_size, input_shapes):
+def generate_input_data(data_size: int, input_shapes: List[VariableShape]) -> List[NDFloat32Array]:
     """Random input data for training."""
     return [generate_random_data(data_size, input_shape)
             for input_shape in input_shapes]
 
 
-def generate_integer_random_data(data_size, low, high, shape):
+def generate_integer_random_data(data_size: int, low: int, high: int, shape: Shape) -> NDUInt32Array:
     """Random data for training."""
     return np.random.randint(
         low=low, high=high, size=get_shape_for_random_data(data_size, replace_none_with(42, shape)))
 
 
-def generate_integer_input_data(data_size, low, highs, input_shapes):
+def generate_integer_input_data(data_size: int, low: int, highs: List[int], input_shapes: List[Shape]) -> List[
+    NDUInt32Array]:
     """Random input data for training."""
     return [generate_integer_random_data(data_size, low, high, input_shape)
             for high, input_shape in zip(highs, input_shapes)]
 
 
-def as_list(value_or_values):
+def as_list(value_or_values: Union[NDFloat32Array, List[NDFloat32Array]]) -> List[NDFloat32Array]:
     """Leave lists untouched, convert non-list types to a singleton list"""
     if isinstance(value_or_values, list):
         return value_or_values
     return [value_or_values]
 
 
-def generate_output_data(data_size, outputs):
+def generate_output_data(data_size: int, outputs: List[NDFloat32Array]) -> List[NDFloat32Array]:
     """Random output data for training."""
     return [generate_random_data(data_size, output.shape[1:])
             for output in as_list(outputs)]
 
 
-def get_test_model_exhaustive():
+def get_test_model_exhaustive() -> Model:
     """Returns a exhaustive test model."""
-    input_shapes = [
+    input_shapes: List[VariableShape] = [
         (2, 3, 4, 5, 6),  # 0
         (2, 3, 4, 5, 6),
         (7, 8, 9, 10),
@@ -155,8 +161,43 @@ def get_test_model_exhaustive():
 
     outputs = []
 
+    outputs.append(Conv1DTranspose(4, 3, padding='valid', use_bias=False)(inputs[6]))
+    outputs.append(Conv2DTranspose(4, (5, 3), padding='valid', use_bias=False)(inputs[4]))
+
+    outputs.append(Conv1DTranspose(4, 1, padding='valid', use_bias=False)(inputs[6]))
+    outputs.append(Conv1DTranspose(4, 1, padding='same', use_bias=False)(inputs[6]))
+    outputs.append(Conv1DTranspose(1, 3, padding='valid', use_bias=False)(inputs[6]))
+    outputs.append(Conv1DTranspose(2, 3, padding='same')(inputs[6]))
+    outputs.append(Conv1DTranspose(2, 5, padding='same', strides=2)(inputs[6]))
+    outputs.append(Conv1DTranspose(2, 5, padding='valid', strides=2)(inputs[6]))
+    # Current CPU implementations do not yet support dilation rates larger than 1
+    # https://github.com/keras-team/keras/issues/20408
+    # https://github.com/keras-team/keras/pull/20737
+    #outputs.append(Conv1DTranspose(3, 5, padding='same', dilation_rate=2)(inputs[6]))
+    #outputs.append(Conv1DTranspose(3, 5, padding='valid', dilation_rate=2)(inputs[6]))
+
+    outputs.append(Conv2DTranspose(4, (1, 1), padding='valid', use_bias=False)(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (1, 1), padding='same', use_bias=False)(inputs[4]))
+    outputs.append(Conv2DTranspose(1, (3, 3))(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (3, 3), padding='valid', use_bias=False)(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (3, 3), padding='same')(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (5, 5), padding='same', strides=(2, 3))(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (5, 5), padding='valid', strides=(2, 3))(inputs[4]))
+    #outputs.append(Conv2DTranspose(4, (5, 5), padding='same', dilation_rate=(2, 3))(inputs[4]))
+    #outputs.append(Conv2DTranspose(4, (5, 5), padding='valid', dilation_rate=(2, 3))(inputs[4]))
+
+    outputs.append(Conv1DTranspose(1, 3, padding='valid')(inputs[6]))
+    outputs.append(Conv1DTranspose(2, 1, padding='same')(inputs[6]))
+    #outputs.append(Conv1DTranspose(3, 4, padding='same', dilation_rate=2)(inputs[6]))
+
+    outputs.append(Conv2DTranspose(4, (3, 3))(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (3, 3), use_bias=False)(inputs[4]))
+    outputs.append(Conv2DTranspose(4, (2, 4), strides=(2, 2), padding='same')(inputs[4]))
+    #outputs.append(Conv2DTranspose(4, (2, 4), padding='same', dilation_rate=(2, 2))(inputs[4]))
+
     outputs.append(Conv1D(1, 3, padding='valid')(inputs[6]))
     outputs.append(Conv1D(2, 1, padding='same')(inputs[6]))
+    outputs.append(Conv1D(2, 1, padding='same', strides=2)(inputs[6]))
     outputs.append(Conv1D(3, 4, padding='causal', dilation_rate=2)(inputs[6]))
     outputs.append(ZeroPadding1D(2)(inputs[6]))
     outputs.append(Cropping1D((2, 3))(inputs[6]))
@@ -170,23 +211,19 @@ def get_test_model_exhaustive():
     outputs.append(GlobalAveragePooling1D(keepdims=True)(inputs[6]))
 
     outputs.append(Normalization(axis=None, mean=2.1, variance=2.2)(inputs[4]))
-    outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[6]))
+    # outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[6]))  # No longer supported in TensorFlow 2.16
     outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[46]))
     outputs.append(Normalization(axis=1, mean=2.1, variance=2.2)(inputs[46]))
     outputs.append(Normalization(axis=-1, mean=2.1, variance=2.2)(inputs[47]))
-    outputs.append(Normalization(axis=1, mean=2.1, variance=2.2)(inputs[47]))
+    # outputs.append(Normalization(axis=1, mean=2.1, variance=2.2)(inputs[47]))  # No longer supported in TensorFlow 2.16
     outputs.append(Normalization(axis=2, mean=2.1, variance=2.2)(inputs[47]))
     for axis in range(1, 6):
-        shape = input_shapes[0][axis - 1]
-        outputs.append(Normalization(axis=axis,
-                                     mean=np.random.rand(shape),
-                                     variance=np.random.rand(shape)
-                                     )(inputs[0]))
+        outputs.append(Normalization(axis=axis, mean=4.2, variance=2.3)(inputs[0]))
 
     outputs.append(Rescaling(23.5, 42.1)(inputs[0]))
 
     outputs.append(Conv2D(4, (3, 3))(inputs[4]))
-    outputs.append(Conv2D(4, (3, 3), use_bias=False)(inputs[4]))
+    outputs.append(Conv2D(4, (3, 3), use_bias=False, padding='valid')(inputs[4]))
     outputs.append(Conv2D(4, (2, 4), strides=(2, 3), padding='same')(inputs[4]))
     outputs.append(Conv2D(4, (2, 4), padding='same', dilation_rate=(2, 3))(inputs[4]))
 
@@ -225,8 +262,8 @@ def get_test_model_exhaustive():
     outputs.append(Resizing(5, 6)(inputs[4]))
     outputs.append(Resizing(19, 53, interpolation="bilinear")(inputs[23]))
     outputs.append(Resizing(19, 53, interpolation="nearest")(inputs[23]))
-    outputs.append(Resizing(7, 9, interpolation="area")(inputs[22]))
-    outputs.append(Resizing(19, 53, interpolation="area")(inputs[23]))
+    # outputs.append(Resizing(7, 9, interpolation="area")(inputs[22]))  # No longer supported in TensorFlow 2.16
+    # outputs.append(Resizing(19, 53, interpolation="area")(inputs[23]))  # No longer supported in TensorFlow 2.16
     outputs.append(Resizing(19, 53, crop_to_aspect_ratio=True)(inputs[23]))
 
     outputs.append(Permute((3, 4, 1, 5, 2))(inputs[0]))
@@ -363,14 +400,15 @@ def get_test_model_exhaustive():
     outputs.append(Minimum()([inputs[6], inputs[7]]))
     outputs.append(Minimum()([inputs[8], inputs[9]]))
 
-    for normalize in [True, False]:
-        outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[8], inputs[9]]))
-        outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[0], inputs[10]]))
-        outputs.append(Dot(axes=1, normalize=normalize)([inputs[0], inputs[10]]))
-        outputs.append(Dot(axes=(3, 1), normalize=normalize)([inputs[31], inputs[32]]))
-        outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[31], inputs[32]]))
-        outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[14], inputs[16]]))
-        outputs.append(Dot(axes=(3, 2), normalize=normalize)([inputs[24], inputs[26]]))
+    # No longer works in TensorFlow 2.16, see: https://github.com/tensorflow/tensorflow/issues/65056
+    # for normalize in [True, False]:
+    # outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[8], inputs[9]]))
+    # outputs.append(Dot(axes=(1, 1), normalize=normalize)([inputs[0], inputs[10]]))
+    # outputs.append(Dot(axes=1, normalize=normalize)([inputs[0], inputs[10]]))
+    # outputs.append(Dot(axes=(3, 1), normalize=normalize)([inputs[31], inputs[32]]))
+    # outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[31], inputs[32]]))
+    # outputs.append(Dot(axes=(2, 3), normalize=normalize)([inputs[14], inputs[16]]))
+    # outputs.append(Dot(axes=(3, 2), normalize=normalize)([inputs[24], inputs[26]]))
 
     outputs.append(Reshape((16,))(inputs[8]))
     outputs.append(Reshape((2, 8))(inputs[8]))
@@ -392,14 +430,14 @@ def get_test_model_exhaustive():
         outputs.append(Concatenate(axis=axis)([inputs[6], inputs[7]]))
     for axis in [-1, 1]:
         outputs.append(Concatenate(axis=axis)([inputs[8], inputs[9]]))
-    for axis in [-1, 2]:
+    for axis in [-1]:  # [-1, 2] no longer supported in TensorFlow 2.16
         outputs.append(Concatenate(axis=axis)([inputs[14], inputs[15]]))
     for axis in [-1, 3]:
         outputs.append(Concatenate(axis=axis)([inputs[16], inputs[17]]))
-    for axis in [-1, 4]:
-        outputs.append(Concatenate(axis=axis)([inputs[18], inputs[19]]))
-    for axis in [-1, 5]:
-        outputs.append(Concatenate(axis=axis)([inputs[20], inputs[21]]))
+    # for axis in [-1, 4]:
+    # outputs.append(Concatenate(axis=axis)([inputs[18], inputs[19]]))  # no longer supported in TensorFlow 2.16
+    # for axis in [-1, 5]:
+    # outputs.append(Concatenate(axis=axis)([inputs[20], inputs[21]]))  # no longer supported in TensorFlow 2.16
 
     outputs.append(UpSampling1D(size=2)(inputs[6]))
     # outputs.append(UpSampling1D(size=2)(inputs[8])) # ValueError: Input 0 of layer up_sampling1d_1 is incompatible with the layer: expected ndim=3, found ndim=2. Full shape received: [None, 16]
@@ -502,39 +540,44 @@ def get_test_model_exhaustive():
     outputs.append(Maximum()([inputs[26], inputs[30], inputs[30]]))
     outputs.append(Concatenate()([inputs[26], inputs[30], inputs[30]]))
 
-    intermediate_input_shape = (3,)
-    intermediate_in = Input(intermediate_input_shape)
-    intermediate_x = intermediate_in
-    intermediate_x = Dense(8)(intermediate_x)
-    intermediate_x = Dense(5, name='duplicate_layer_name')(intermediate_x)
-    intermediate_model = Model(
-        inputs=[intermediate_in], outputs=[intermediate_x],
-        name='intermediate_model')
-    intermediate_model.compile(loss='mse', optimizer='nadam')
+    # TensorFlow 2.16 no longer puts
+    # "inbound_nodes": []
+    # for such nested models.
+    # todo: Check if the situation resolved with later versions.
+    if False:
+        intermediate_input_shape = (3,)
+        intermediate_in = Input(intermediate_input_shape)
+        intermediate_x = intermediate_in
+        intermediate_x = Dense(8)(intermediate_x)
+        intermediate_x = Dense(5, name='duplicate_layer_name')(intermediate_x)
+        intermediate_model = Model(
+            inputs=[intermediate_in], outputs=[intermediate_x],
+            name='intermediate_model')
+        intermediate_model.compile(loss='mse', optimizer='nadam')
 
-    x = intermediate_model(x)  # (1, 1, 5)
+        x = intermediate_model(x)[0]  # (1, 1, 5)
 
-    intermediate_model_2 = Sequential()
-    intermediate_model_2.add(Dense(7, input_shape=(5,)))
-    intermediate_model_2.add(Dense(5, name='duplicate_layer_name'))
-    intermediate_model_2.compile(optimizer='rmsprop',
-                                 loss='categorical_crossentropy')
+        intermediate_model_2 = Sequential(name="intermediate_model_2")
+        intermediate_model_2.add(Dense(7, input_shape=(5,)))
+        intermediate_model_2.add(Dense(5, name='duplicate_layer_name'))
+        intermediate_model_2.compile(optimizer='rmsprop',
+                                     loss='categorical_crossentropy')
 
-    x = intermediate_model_2(x)  # (1, 1, 5)
+        x = intermediate_model_2(x)  # (1, 1, 5)
 
-    intermediate_model_3_nested = Sequential()
-    intermediate_model_3_nested.add(Dense(7, input_shape=(6,)))
-    intermediate_model_3_nested.compile(optimizer='rmsprop', loss='categorical_crossentropy')
+        intermediate_model_3_nested = Sequential(name="intermediate_model_3_nested")
+        intermediate_model_3_nested.add(Dense(7, input_shape=(6,)))
+        intermediate_model_3_nested.compile(optimizer='rmsprop', loss='categorical_crossentropy')
 
-    intermediate_model_3 = Sequential()
-    intermediate_model_3.add(Dense(6, input_shape=(5,)))
-    intermediate_model_3.add(intermediate_model_3_nested)
-    intermediate_model_3.add(Dense(8))
-    intermediate_model_3.compile(optimizer='rmsprop', loss='categorical_crossentropy')
+        intermediate_model_3 = Sequential(name="intermediate_model_3")
+        intermediate_model_3.add(Dense(6, input_shape=(5,)))
+        intermediate_model_3.add(intermediate_model_3_nested)
+        intermediate_model_3.add(Dense(8))
+        intermediate_model_3.compile(optimizer='rmsprop', loss='categorical_crossentropy')
 
-    x = intermediate_model_3(x)  # (1, 1, 8)
+        x = intermediate_model_3(x)  # (1, 1, 8)
 
-    x = Dense(3)(x)  # (1, 1, 3)
+        x = Dense(3)(x)  # (1, 1, 3)
 
     shared_activation = Activation('tanh')
 
@@ -551,7 +594,7 @@ def get_test_model_exhaustive():
         Activation('exponential')(inputs[25]),
         Activation('gelu')(inputs[25]),
         Activation('softsign')(inputs[25]),
-        LeakyReLU()(inputs[25]),
+        LeakyReLU(negative_slope=0.5)(inputs[25]),
         ReLU()(inputs[25]),
         ReLU(max_value=0.4, negative_slope=1.1, threshold=0.3)(inputs[25]),
         ELU()(inputs[25]),
@@ -577,15 +620,15 @@ def get_test_model_exhaustive():
     return model
 
 
-def get_test_model_embedding():
-    """Returns a minimalistic test model for the embedding layer."""
+def get_test_model_embedding() -> Model:
+    """Returns a minimalistic test model for the Embedding and CategoryEncoding layers."""
 
     input_dims = [
         1023,  # maximum integer value in input data
         255,
         15,
     ]
-    input_shapes = [
+    input_shapes: list[tuple[int, ...]] = [
         (100,),  # must be single-element tuple (for sequence length)
         (1000,),
         (1,),
@@ -598,17 +641,13 @@ def get_test_model_embedding():
     outputs = []
     for k in range(2):
         embedding = Embedding(input_dim=input_dims[k], output_dim=output_dims[k])(inputs[k])
-        lstm = LSTM(
-            units=4,
-            recurrent_activation='sigmoid',
-            return_sequences=False
-        )(embedding)
-
-        outputs.append(lstm)
+        outputs.append(embedding)
 
     outputs.append(CategoryEncoding(1024, output_mode='multi_hot', sparse=False)(inputs[0]))
-    outputs.append(CategoryEncoding(1024, output_mode='count', sparse=False)(inputs[0]))
-    outputs.append(CategoryEncoding(16, output_mode='one_hot', sparse=False)(inputs[2]))
+    # No longer working since TF 2.16: https://github.com/tensorflow/tensorflow/issues/65390
+    # Error: Value passed to parameter 'values' has DataType float32 not in list of allowed values: int32, int64
+    # outputs.append(CategoryEncoding(1024, output_mode='count', sparse=False)(inputs[0]))
+    # outputs.append(CategoryEncoding(16, output_mode='one_hot', sparse=False)(inputs[2]))
     # Error: Value passed to parameter 'values' has DataType float32 not in list of allowed values: int32, int64
     # outputs.append(CategoryEncoding(1023, output_mode='multi_hot', sparse=True)(inputs[0]))
 
@@ -624,266 +663,7 @@ def get_test_model_embedding():
     return model
 
 
-def get_test_model_recurrent():
-    """Returns a minimalistic test model for recurrent layers."""
-    input_shapes = [
-        (17, 4),
-        (1, 10),
-        (20, 40),
-        (6, 7, 10, 3)
-    ]
-
-    outputs = []
-
-    inputs = [Input(shape=s) for s in input_shapes]
-
-    inp = PReLU()(inputs[0])
-
-    lstm = Bidirectional(LSTM(units=4,
-                              return_sequences=True,
-                              bias_initializer='random_uniform',  # default is zero use random to test computation
-                              activation='tanh',
-                              recurrent_activation='relu'), merge_mode='concat')(inp)
-
-    lstm2 = Bidirectional(LSTM(units=6,
-                               return_sequences=True,
-                               bias_initializer='random_uniform',
-                               activation='elu',
-                               recurrent_activation='hard_sigmoid'), merge_mode='sum')(lstm)
-
-    lstm3 = LSTM(units=10,
-                 return_sequences=False,
-                 bias_initializer='random_uniform',
-                 activation='selu',
-                 recurrent_activation='sigmoid')(lstm2)
-
-    outputs.append(lstm3)
-
-    conv1 = Conv1D(2, 1, activation='sigmoid')(inputs[1])
-    lstm4 = LSTM(units=15,
-                 return_sequences=False,
-                 bias_initializer='random_uniform',
-                 activation='tanh',
-                 recurrent_activation='elu')(conv1)
-
-    dense = (Dense(23, activation='sigmoid'))(lstm4)
-    outputs.append(dense)
-
-    time_dist_1 = TimeDistributed(Conv2D(2, (3, 3), use_bias=True))(inputs[3])
-    flatten_1 = TimeDistributed(Flatten())(time_dist_1)
-
-    outputs.append(Bidirectional(LSTM(units=6,
-                                      return_sequences=True,
-                                      bias_initializer='random_uniform',
-                                      activation='tanh',
-                                      recurrent_activation='sigmoid'), merge_mode='ave')(flatten_1))
-
-    outputs.append(TimeDistributed(MaxPooling2D(2, 2))(inputs[3]))
-    outputs.append(TimeDistributed(AveragePooling2D(2, 2))(inputs[3]))
-    outputs.append(TimeDistributed(BatchNormalization())(inputs[3]))
-
-    nested_inputs = Input(shape=input_shapes[0][1:])
-    nested_x = Dense(5, activation='relu')(nested_inputs)
-    nested_predictions = Dense(3, activation='softmax')(nested_x)
-    nested_model = Model(inputs=nested_inputs, outputs=nested_predictions)
-    nested_model.compile(loss='categorical_crossentropy', optimizer='nadam')
-    outputs.append(TimeDistributed(nested_model)(inputs[0]))
-
-    nested_sequential_model = Sequential()
-    nested_sequential_model.add(Flatten(input_shape=input_shapes[0][1:]))
-    nested_sequential_model.compile(optimizer='rmsprop',
-                                    loss='categorical_crossentropy')
-    outputs.append(TimeDistributed(nested_sequential_model)(inputs[0]))
-
-    model = Model(inputs=inputs, outputs=outputs, name='test_model_recurrent')
-    model.compile(loss='mse', optimizer='nadam')
-
-    # fit to dummy data
-    training_data_size = 2
-    data_in = generate_input_data(training_data_size, input_shapes)
-    initial_data_out = model.predict(data_in)
-    data_out = generate_output_data(training_data_size, initial_data_out)
-    model.fit(data_in, data_out, epochs=10)
-    return model
-
-
-def get_test_model_lstm():
-    """Returns a test model for Long Short-Term Memory (LSTM) layers."""
-
-    input_shapes = [
-        (17, 4),
-        (1, 10),
-        (None, 4),
-        (12,),
-        (12,)
-    ]
-    inputs = [Input(shape=s) for s in input_shapes]
-    outputs = []
-
-    for inp in inputs[:2]:
-        lstm_sequences = LSTM(
-            units=8,
-            recurrent_activation='relu',
-            return_sequences=True
-        )(inp)
-        lstm_regular = LSTM(
-            units=3,
-            recurrent_activation='sigmoid',
-            return_sequences=False
-        )(lstm_sequences)
-        outputs.append(lstm_regular)
-        lstm_state, state_h, state_c = LSTM(
-            units=3,
-            recurrent_activation='sigmoid',
-            return_state=True
-        )(inp)
-        outputs.append(lstm_state)
-        outputs.append(state_h)
-        outputs.append(state_c)
-
-        lstm_bidi_sequences = Bidirectional(
-            LSTM(
-                units=4,
-                recurrent_activation='hard_sigmoid',
-                return_sequences=True
-            )
-        )(inp)
-        lstm_bidi = Bidirectional(
-            LSTM(
-                units=6,
-                recurrent_activation='linear',
-                return_sequences=False
-            )
-        )(lstm_bidi_sequences)
-        outputs.append(lstm_bidi)
-
-        lstm_gpu_regular = LSTM(
-            units=3,
-            activation='tanh',
-            recurrent_activation='sigmoid',
-            use_bias=True
-        )(inp)
-
-        lstm_gpu_bidi = Bidirectional(
-            LSTM(
-                units=3,
-                activation='tanh',
-                recurrent_activation='sigmoid',
-                use_bias=True
-            )
-        )(inp)
-    outputs.append(lstm_gpu_regular)
-    outputs.append(lstm_gpu_bidi)
-
-    outputs.extend(LSTM(units=12, return_sequences=True,
-                        return_state=True)(inputs[2], initial_state=[inputs[3], inputs[4]]))
-
-    model = Model(inputs=inputs, outputs=outputs, name='test_model_lstm')
-    model.compile(loss='mse', optimizer='nadam')
-
-    # fit to dummy data
-    training_data_size = 2
-    data_in = generate_input_data(training_data_size, input_shapes)
-    initial_data_out = model.predict(data_in)
-    data_out = generate_output_data(training_data_size, initial_data_out)
-    model.fit(data_in, data_out, epochs=10)
-    return model
-
-
-def get_test_model_gru():
-    return get_test_model_gru_stateful_optional(False)
-
-
-def get_test_model_gru_stateful():
-    return get_test_model_gru_stateful_optional(True)
-
-
-def get_test_model_gru_stateful_optional(stateful):
-    """Returns a test model for Gated Recurrent Unit (GRU) layers."""
-    input_shapes = [
-        (17, 4),
-        (1, 10)
-    ]
-    stateful_batch_size = 1
-    inputs = [Input(batch_shape=(stateful_batch_size,) + s) for s in input_shapes]
-    outputs = []
-
-    for inp in inputs:
-        gru_sequences = GRU(
-            stateful=stateful,
-            units=8,
-            recurrent_activation='relu',
-            reset_after=True,
-            return_sequences=True,
-            use_bias=True
-        )(inp)
-        gru_regular = GRU(
-            stateful=stateful,
-            units=3,
-            recurrent_activation='sigmoid',
-            reset_after=True,
-            return_sequences=False,
-            use_bias=False
-        )(gru_sequences)
-        outputs.append(gru_regular)
-
-        gru_bidi_sequences = Bidirectional(
-            GRU(
-                stateful=stateful,
-                units=4,
-                recurrent_activation='hard_sigmoid',
-                reset_after=False,
-                return_sequences=True,
-                use_bias=True
-            )
-        )(inp)
-        gru_bidi = Bidirectional(
-            GRU(
-                stateful=stateful,
-                units=6,
-                recurrent_activation='sigmoid',
-                reset_after=True,
-                return_sequences=False,
-                use_bias=False
-            )
-        )(gru_bidi_sequences)
-        outputs.append(gru_bidi)
-
-        gru_gpu_regular = GRU(
-            stateful=stateful,
-            units=3,
-            activation='tanh',
-            recurrent_activation='sigmoid',
-            reset_after=True,
-            use_bias=True
-        )(inp)
-
-        gru_gpu_bidi = Bidirectional(
-            GRU(
-                stateful=stateful,
-                units=3,
-                activation='tanh',
-                recurrent_activation='sigmoid',
-                reset_after=True,
-                use_bias=True
-            )
-        )(inp)
-        outputs.append(gru_gpu_regular)
-        outputs.append(gru_gpu_bidi)
-
-    model = Model(inputs=inputs, outputs=outputs, name='test_model_gru')
-    model.compile(loss='mse', optimizer='nadam')
-
-    # fit to dummy data
-    training_data_size = stateful_batch_size
-    data_in = generate_input_data(training_data_size, input_shapes)
-    initial_data_out = model.predict(data_in)
-    data_out = generate_output_data(training_data_size, initial_data_out)
-    model.fit(data_in, data_out, batch_size=stateful_batch_size, epochs=10)
-    return model
-
-
-def get_test_model_variable():
+def get_test_model_variable() -> Model:
     """Returns a model with variably shaped input tensors."""
 
     input_shapes = [
@@ -904,7 +684,7 @@ def get_test_model_variable():
     outputs.append(Reshape((2, -1))(inputs[2]))
     outputs.append(Reshape((-1, 2))(inputs[2]))
     outputs.append(MaxPooling2D()(inputs[1]))
-    outputs.append(AveragePooling1D()(inputs[2]))
+    outputs.append(AveragePooling1D(2)(inputs[2]))
 
     outputs.append(PReLU(shared_axes=[1, 2])(inputs[0]))
     outputs.append(PReLU(shared_axes=[1, 2])(inputs[1]))
@@ -923,7 +703,25 @@ def get_test_model_variable():
     return model
 
 
-def get_test_model_sequential():
+def get_test_model_autoencoder() -> Model:
+    """Returns a minimal autoencoder test model."""
+    input_img = Input(shape=(1,), name='input_img')
+    encoded = Identity()(input_img)  # Since it's about testing node connections, this suffices.
+    encoder = Model(input_img, encoded, name="encoder")
+
+    input_encoded = Input(shape=(1,), name='input_encoded')
+    decoded = Identity()(input_encoded)
+    decoder = Model(input_encoded, decoded, name="decoder")
+
+    autoencoder_input = Input(shape=(1,), name='input_autoencoder')
+    x = encoder(autoencoder_input)
+    autoencodedanddecoded = decoder(x)
+    autoencoder = Model(inputs=autoencoder_input, outputs=autoencodedanddecoded, name="autoencoder")
+    autoencoder.compile(optimizer='sgd', loss='mse')
+    return autoencoder
+
+
+def get_test_model_sequential() -> Model:
     """Returns a typical (VGG-like) sequential test model."""
     model = Sequential()
     model.add(Conv2D(8, (3, 3), activation='relu', input_shape=(32, 32, 3)))
@@ -955,93 +753,10 @@ def get_test_model_sequential():
     return model
 
 
-def get_test_model_lstm_stateful():
-    stateful_batch_size = 1
-    input_shapes = [
-        (17, 4),
-        (1, 10),
-        (None, 4),
-        (12,),
-        (12,)
-    ]
-
-    inputs = [Input(batch_shape=(stateful_batch_size,) + s) for s in input_shapes]
-    outputs = []
-    for in_num, inp in enumerate(inputs[:2]):
-        stateful = bool((in_num + 1) % 2)
-        lstm_sequences = LSTM(
-            stateful=stateful,
-            units=8,
-            recurrent_activation='relu',
-            return_sequences=True,
-            name='lstm_sequences_' + str(in_num) + '_st-' + str(stateful)
-        )(inp)
-        stateful = bool((in_num) % 2)
-        lstm_regular = LSTM(
-            stateful=stateful,
-            units=3,
-            recurrent_activation='sigmoid',
-            return_sequences=False,
-            name='lstm_regular_' + str(in_num) + '_st-' + str(stateful)
-        )(lstm_sequences)
-        outputs.append(lstm_regular)
-        stateful = bool((in_num + 1) % 2)
-        lstm_state, state_h, state_c = LSTM(
-            stateful=stateful,
-            units=3,
-            recurrent_activation='sigmoid',
-            return_state=True,
-            name='lstm_state_return_' + str(in_num) + '_st-' + str(stateful)
-        )(inp)
-        outputs.append(lstm_state)
-        outputs.append(state_h)
-        outputs.append(state_c)
-        stateful = bool((in_num + 1) % 2)
-        lstm_bidi_sequences = Bidirectional(
-            LSTM(
-                stateful=stateful,
-                units=4,
-                recurrent_activation='hard_sigmoid',
-                return_sequences=True,
-                name='bi-lstm1_' + str(in_num) + '_st-' + str(stateful)
-            )
-        )(inp)
-        stateful = bool((in_num) % 2)
-        lstm_bidi = Bidirectional(
-            LSTM(
-                stateful=stateful,
-                units=6,
-                recurrent_activation='linear',
-                return_sequences=False,
-                name='bi-lstm2_' + str(in_num) + '_st-' + str(stateful)
-            )
-        )(lstm_bidi_sequences)
-        outputs.append(lstm_bidi)
-
-    initial_state_stateful = LSTM(units=12, return_sequences=True, stateful=True, return_state=True,
-                                  name='initial_state_stateful')(inputs[2], initial_state=[inputs[3], inputs[4]])
-    outputs.extend(initial_state_stateful)
-    initial_state_not_stateful = LSTM(units=12, return_sequences=False, stateful=False, return_state=True,
-                                      name='initial_state_not_stateful')(inputs[2],
-                                                                         initial_state=[inputs[3], inputs[4]])
-    outputs.extend(initial_state_not_stateful)
-    model = Model(inputs=inputs, outputs=outputs)
-    model.compile(loss='mean_squared_error', optimizer='nadam')
-
-    # fit to dummy data
-    training_data_size = stateful_batch_size
-    data_in = generate_input_data(training_data_size, input_shapes)
-    initial_data_out = model.predict(data_in)
-    data_out = generate_output_data(training_data_size, initial_data_out)
-
-    model.fit(data_in, data_out, batch_size=stateful_batch_size, epochs=10)
-    return model
-
-
-def main():
+def main() -> None:
     """Generate different test models and save them to the given directory."""
     if len(sys.argv) != 3:
-        print('usage: [model name] [destination file path]')
+        print('usage: [model name] [destination file path]', flush=True)
         sys.exit(1)
     else:
         model_name = sys.argv[1]
@@ -1050,13 +765,9 @@ def main():
         get_model_functions = {
             'exhaustive': get_test_model_exhaustive,
             'embedding': get_test_model_embedding,
-            'recurrent': get_test_model_recurrent,
-            'lstm': get_test_model_lstm,
-            'gru': get_test_model_gru,
             'variable': get_test_model_variable,
+            'autoencoder': get_test_model_autoencoder,
             'sequential': get_test_model_sequential,
-            'lstm_stateful': get_test_model_lstm_stateful,
-            'gru_stateful': get_test_model_gru_stateful
         }
 
         if not model_name in get_model_functions:
@@ -1073,7 +784,7 @@ def main():
         # see https://github.com/fchollet/keras/issues/7682
         model = load_model(dest_path)
         model.summary()
-        # plot_model(model, to_file= str(model_name) + '.png', show_shapes=True, show_layer_names=True)  #### DEBUG stateful
+        # plot_model(model, to_file= str(model_name) + '.png', show_shapes=True, show_layer_names=True)
 
 
 if __name__ == "__main__":
